import re
import unicodedata
import urllib
import doctest

ar_chars = """\
ء	ARABIC LETTER HAMZA
آ	ARABIC LETTER ALEF WITH MADDA ABOVE
أ	ARABIC LETTER ALEF WITH HAMZA ABOVE
ؤ	ARABIC LETTER WAW WITH HAMZA ABOVE
إ	ARABIC LETTER ALEF WITH HAMZA BELOW
ئ	ARABIC LETTER YEH WITH HAMZA ABOVE
ا	ARABIC LETTER ALEF
ب	ARABIC LETTER BEH
ة	ARABIC LETTER TEH MARBUTA
ت	ARABIC LETTER TEH
ث	ARABIC LETTER THEH
ج	ARABIC LETTER JEEM
ح	ARABIC LETTER HAH
خ	ARABIC LETTER KHAH
د	ARABIC LETTER DAL
ذ	ARABIC LETTER THAL
ر	ARABIC LETTER REH
ز	ARABIC LETTER ZAIN
س	ARABIC LETTER SEEN
ش	ARABIC LETTER SHEEN
ص	ARABIC LETTER SAD
ض	ARABIC LETTER DAD
ط	ARABIC LETTER TAH
ظ	ARABIC LETTER ZAH
ع	ARABIC LETTER AIN
غ	ARABIC LETTER GHAIN
ـ	ARABIC TATWEEL
ف	ARABIC LETTER FEH
ق	ARABIC LETTER QAF
ك	ARABIC LETTER KAF
ل	ARABIC LETTER LAM
م	ARABIC LETTER MEEM
ن	ARABIC LETTER NOON
ه	ARABIC LETTER HEH
و	ARABIC LETTER WAW
ى	ARABIC LETTER ALEF MAKSURA
ي	ARABIC LETTER YEH
ً	ARABIC FATHATAN
ٌ	ARABIC DAMMATAN
ٍ	ARABIC KASRATAN
َ	ARABIC FATHA
ُ	ARABIC DAMMA
ِ	ARABIC KASRA
ّ	ARABIC SHADDA
ْ	ARABIC SUKUN
٠	ARABIC-INDIC DIGIT ZERO
١	ARABIC-INDIC DIGIT ONE
٢	ARABIC-INDIC DIGIT TWO
٣	ARABIC-INDIC DIGIT THREE
٤	ARABIC-INDIC DIGIT FOUR
٥	ARABIC-INDIC DIGIT FIVE
٦	ARABIC-INDIC DIGIT SIX
٧	ARABIC-INDIC DIGIT SEVEN
٨	ARABIC-INDIC DIGIT EIGHT
٩	ARABIC-INDIC DIGIT NINE
ٮ	ARABIC LETTER DOTLESS BEH
ٰ	ARABIC LETTER SUPERSCRIPT ALEF
ٹ	ARABIC LETTER TTEH
پ	ARABIC LETTER PEH
چ	ARABIC LETTER TCHEH
ژ	ARABIC LETTER JEH
ک	ARABIC LETTER KEHEH
گ	ARABIC LETTER GAF
ی	ARABIC LETTER FARSI YEH
ے	ARABIC LETTER YEH BARREE
۱	EXTENDED ARABIC-INDIC DIGIT ONE
۲	EXTENDED ARABIC-INDIC DIGIT TWO
۳	EXTENDED ARABIC-INDIC DIGIT THREE
۴	EXTENDED ARABIC-INDIC DIGIT FOUR
۵	EXTENDED ARABIC-INDIC DIGIT FIVE
۶	EXTENDED ARABIC-INDIC DIGIT SIX
۷	EXTENDED ARABIC-INDIC DIGIT SEVEN
۸	EXTENDED ARABIC-INDIC DIGIT EIGHT
۹	EXTENDED ARABIC-INDIC DIGIT NINE
۰	EXTENDED ARABIC-INDIC DIGIT ZERO
"""
##‌	ZERO WIDTH NON-JOINER
##‍	ZERO WIDTH JOINER"""
ar_chars = [x.split("\t")[0] for x in ar_chars.splitlines()]
#ar_chars = "ذ١٢٣٤٥٦٧٨٩٠ّـضصثقفغعهخحجدًٌَُلإإشسيبلاتنمكطٍِلأأـئءؤرلاىةوزظْلآآ"
ar_char = re.compile("[{}]".format("".join(ar_chars))) # regex for one Arabic character
ar_tok = re.compile("[{}]+".format("".join(ar_chars))) # regex for one Arabic token
noise = re.compile(""" ّ    | # Tashdīd / Shadda
                       َ    | # Fatḥa
                       ً    | # Tanwīn Fatḥ / Fatḥatān
                       ُ    | # Ḍamma
                       ٌ    | # Tanwīn Ḍamm / Ḍammatān
                       ِ    | # Kasra
                       ٍ    | # Tanwīn Kasr / Kasratān
                       ْ    | # Sukūn
                       ۡ    | # Quranic Sukūn
                       ࣰ    | # Quranic Open Fatḥatān
                       ࣱ    | # Quranic Open Ḍammatān
                       ࣲ    | # Quranic Open Kasratān
                       ٰ    | # Dagger Alif
                       ـ   | # Taṭwīl / Kashīda
                       ٖ    | # ARABIC SUBSCRIPT ALEF
                       ٗ    | # ARABIC INVERTED DAMMA
                       ۡ    | # ARABIC SMALL HIGH DOTLESS HEAD OF KHAH = Qur'anic sukūn
                       ۤ      # ARABIC SMALL HIGH MADDA
                   """, re.VERBOSE)

# Whitelist of characters that are allowed in OpenITI texts:

allowed_chars = """\
ء	ARABIC LETTER HAMZA
آ	ARABIC LETTER ALEF WITH MADDA ABOVE
أ	ARABIC LETTER ALEF WITH HAMZA ABOVE
ؤ	ARABIC LETTER WAW WITH HAMZA ABOVE
إ	ARABIC LETTER ALEF WITH HAMZA BELOW
ئ	ARABIC LETTER YEH WITH HAMZA ABOVE
ا	ARABIC LETTER ALEF
ب	ARABIC LETTER BEH
ة	ARABIC LETTER TEH MARBUTA
ت	ARABIC LETTER TEH
ث	ARABIC LETTER THEH
ج	ARABIC LETTER JEEM
ح	ARABIC LETTER HAH
خ	ARABIC LETTER KHAH
د	ARABIC LETTER DAL
ذ	ARABIC LETTER THAL
ر	ARABIC LETTER REH
ز	ARABIC LETTER ZAIN
س	ARABIC LETTER SEEN
ش	ARABIC LETTER SHEEN
ص	ARABIC LETTER SAD
ض	ARABIC LETTER DAD
ط	ARABIC LETTER TAH
ظ	ARABIC LETTER ZAH
ع	ARABIC LETTER AIN
غ	ARABIC LETTER GHAIN
ف	ARABIC LETTER FEH
ق	ARABIC LETTER QAF
ك	ARABIC LETTER KAF
ل	ARABIC LETTER LAM
م	ARABIC LETTER MEEM
ن	ARABIC LETTER NOON
ه	ARABIC LETTER HEH
و	ARABIC LETTER WAW
ى	ARABIC LETTER ALEF MAKSURA
ي	ARABIC LETTER YEH
٠	ARABIC-INDIC DIGIT ZERO
١	ARABIC-INDIC DIGIT ONE
٢	ARABIC-INDIC DIGIT TWO
٣	ARABIC-INDIC DIGIT THREE
٤	ARABIC-INDIC DIGIT FOUR
٥	ARABIC-INDIC DIGIT FIVE
٦	ARABIC-INDIC DIGIT SIX
٧	ARABIC-INDIC DIGIT SEVEN
٨	ARABIC-INDIC DIGIT EIGHT
٩	ARABIC-INDIC DIGIT NINE
ٹ	ARABIC LETTER TTEH
پ	ARABIC LETTER PEH
چ	ARABIC LETTER TCHEH
ڈ	ARABIC LETTER DDAL
ڑ	ARABIC LETTER RREH
ژ	ARABIC LETTER JEH
ک	ARABIC LETTER KEHEH
ݣ	ARABIC LETTER KEHEH WITH THREE DOTS ABOVE
ڭ	ARABIC LETTER NG
گ	ARABIC LETTER GAF
ۀ	ARABIC LETTER HEH WITH YEH ABOVE
ہ	ARABIC LETTER HEH GOAL
ۂ	ARABIC LETTER HEH GOAL WITH HAMZA ABOVE
ی	ARABIC LETTER FARSI YEH
ے	ARABIC LETTER YEH BARREE
ۓ	ARABIC LETTER YEH BARREE WITH HAMZA ABOVE
ھ	ARABIC LETTER HEH DOACHASHMEE
ە	ARABIC LETTER AE
ں	ARABIC LETTER NOON GHUNNA
ٮ   ARABIC LETTER DOTLESS BEH
ٯ   ARABIC LETTER DOTLESS QAF
ڡ   ARABIC LETTER DOTLESS FEH
ڣ   ARABIC LETTER FEH WITH DOT BELOW
‌	ZERO WIDTH NON-JOINER
ٔ	ARABIC HAMZA ABOVE (needed for the Farsi izafeh)
#	NUMBER SIGN
%	PERCENT SIGN
(	LEFT PARENTHESIS
)	RIGHT PARENTHESIS
.	FULL STOP
/	SOLIDUS
0	DIGIT ZERO
1	DIGIT ONE
2	DIGIT TWO
3	DIGIT THREE
4	DIGIT FOUR
5	DIGIT FIVE
7	DIGIT SEVEN
8	DIGIT EIGHT
9	DIGIT NINE
:	COLON
|	VERTICAL LINE
~	TILDE
؟	ARABIC QUESTION MARK
،	ARABIC COMMA
!	EXCLAMATION MARK
$	DOLLAR SIGN
*	ASTERISK
-	HYPHEN-MINUS
_	LOW LINE (i.e., underscore)
«	LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
»	RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
؛	ARABIC SEMICOLON
!	EXCLAMATION MARK
"	QUOTATION MARK
,	COMMA
=	EQUALS SIGN
?	QUESTION MARK
“	LEFT DOUBLE QUOTATION MARK
”	RIGHT DOUBLE QUOTATION MARK
’	RIGHT SINGLE QUOTATION MARK
&   AMPERSAND
§   SECTION SIGN
¶	PILCROW SIGN
¬	NOT SIGN
•	BULLET
<	LESS-THAN SIGN
>	GREATER-THAN SIGN
{	LEFT CURLY BRACKET
}	RIGHT CURLY BRACKET
+	PLUS SIGN
;	SEMICOLON
@	COMMERCIAL AT
ְ	HEBREW POINT SHEVA
ֳ	HEBREW POINT HATAF QAMATS
ִ	HEBREW POINT HIRIQ
ֵ	HEBREW POINT TSERE
ֶ	HEBREW POINT SEGOL
ַ	HEBREW POINT PATAH
ָ	HEBREW POINT QAMATS
ֹ	HEBREW POINT HOLAM
ּ	HEBREW POINT DAGESH OR MAPIQ
א	HEBREW LETTER ALEF
ב	HEBREW LETTER BET
ג	HEBREW LETTER GIMEL
ד	HEBREW LETTER DALET
ה	HEBREW LETTER HE
ו	HEBREW LETTER VAV
ז	HEBREW LETTER ZAYIN
ח	HEBREW LETTER HET
ט	HEBREW LETTER TET
י	HEBREW LETTER YOD
ך	HEBREW LETTER FINAL KAF
כ	HEBREW LETTER KAF
ל	HEBREW LETTER LAMED
ם	HEBREW LETTER FINAL MEM
מ	HEBREW LETTER MEM
ן	HEBREW LETTER FINAL NUN
נ	HEBREW LETTER NUN
ס	HEBREW LETTER SAMEKH
ע	HEBREW LETTER AYIN
ף	HEBREW LETTER FINAL PE
פ	HEBREW LETTER PE
ץ	HEBREW LETTER FINAL TSADI
צ	HEBREW LETTER TSADI
ק	HEBREW LETTER QOF
ר	HEBREW LETTER RESH
ש	HEBREW LETTER SHIN
ת	HEBREW LETTER TAV
І	CYRILLIC CAPITAL LETTER BYELORUSSIAN-UKRAINIAN I
А	CYRILLIC CAPITAL LETTER A
Б	CYRILLIC CAPITAL LETTER BE
В	CYRILLIC CAPITAL LETTER VE
Г	CYRILLIC CAPITAL LETTER GHE
Д	CYRILLIC CAPITAL LETTER DE
Е	CYRILLIC CAPITAL LETTER IE
Ж	CYRILLIC CAPITAL LETTER ZHE
З	CYRILLIC CAPITAL LETTER ZE
И	CYRILLIC CAPITAL LETTER I
Й	CYRILLIC CAPITAL LETTER SHORT I
К	CYRILLIC CAPITAL LETTER KA
Л	CYRILLIC CAPITAL LETTER EL
М	CYRILLIC CAPITAL LETTER EM
Н	CYRILLIC CAPITAL LETTER EN
О	CYRILLIC CAPITAL LETTER O
П	CYRILLIC CAPITAL LETTER PE
Р	CYRILLIC CAPITAL LETTER ER
С	CYRILLIC CAPITAL LETTER ES
Т	CYRILLIC CAPITAL LETTER TE
У	CYRILLIC CAPITAL LETTER U
Ф	CYRILLIC CAPITAL LETTER EF
Х	CYRILLIC CAPITAL LETTER HA
Ц	CYRILLIC CAPITAL LETTER TSE
Ч	CYRILLIC CAPITAL LETTER CHE
Ш	CYRILLIC CAPITAL LETTER SHA
Щ	CYRILLIC CAPITAL LETTER SHCHA
Ъ	CYRILLIC CAPITAL LETTER HARD SIGN
Ы	CYRILLIC CAPITAL LETTER YERU
Ь	CYRILLIC CAPITAL LETTER SOFT SIGN
Э	CYRILLIC CAPITAL LETTER E
Ю	CYRILLIC CAPITAL LETTER YU
Я	CYRILLIC CAPITAL LETTER YA
а	CYRILLIC SMALL LETTER A
б	CYRILLIC SMALL LETTER BE
в	CYRILLIC SMALL LETTER VE
г	CYRILLIC SMALL LETTER GHE
д	CYRILLIC SMALL LETTER DE
е	CYRILLIC SMALL LETTER IE
ж	CYRILLIC SMALL LETTER ZHE
з	CYRILLIC SMALL LETTER ZE
и	CYRILLIC SMALL LETTER I
й	CYRILLIC SMALL LETTER SHORT I
к	CYRILLIC SMALL LETTER KA
л	CYRILLIC SMALL LETTER EL
м	CYRILLIC SMALL LETTER EM
н	CYRILLIC SMALL LETTER EN
о	CYRILLIC SMALL LETTER O
п	CYRILLIC SMALL LETTER PE
р	CYRILLIC SMALL LETTER ER
с	CYRILLIC SMALL LETTER ES
т	CYRILLIC SMALL LETTER TE
у	CYRILLIC SMALL LETTER U
ф	CYRILLIC SMALL LETTER EF
х	CYRILLIC SMALL LETTER HA
ц	CYRILLIC SMALL LETTER TSE
ч	CYRILLIC SMALL LETTER CHE
ш	CYRILLIC SMALL LETTER SHA
щ	CYRILLIC SMALL LETTER SHCHA
ъ	CYRILLIC SMALL LETTER HARD SIGN
ы	CYRILLIC SMALL LETTER YERU
ь	CYRILLIC SMALL LETTER SOFT SIGN
э	CYRILLIC SMALL LETTER E
ю	CYRILLIC SMALL LETTER YU
я	CYRILLIC SMALL LETTER YA
і	CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I
Ѣ	CYRILLIC CAPITAL LETTER YAT
ѣ	CYRILLIC SMALL LETTER YAT
Ѳ	CYRILLIC CAPITAL LETTER FITA
"""
allowed_chars = [x.split("\t")[0] for x in allowed_chars.splitlines()]
allowed_chars = [c for c in allowed_chars if c not in ("-", ".")]

# In addition to the above characters, also include Latin script, whitespace and punctuation:
transcription_chars = "0-9a-zA-ZāĀăĂēĒĕĔṭṬṯṮūŪīĪĭĬİıōŌṣṢšŠḍḌḏḎǧǦġĠğĞḫḪḥḤḳḲẓẒžŽčČçÇñÑãÃáÁàÀäÄéÉèÈêÊëËïÏîÎôÔóÓòÒōÕöÖüÜûÛúÚùÙʿʾ' "
escaped_chars = r"\"\n\t\[\]\.\-\\"

# build a regex to match all allowed characters (and all those that are not allowed):
allowed_chars_regex = re.compile(r"[{}{}{}]+".format("".join(allowed_chars), transcription_chars, escaped_chars))
unwanted_chars_regex = re.compile(r"[^{}{}{}]+".format("".join(allowed_chars), transcription_chars, escaped_chars))




def denoise(text):
    """Remove non-consonantal characters from Arabic text.

    Examples:
        >>> denoise("وَالَّذِينَ يُؤْمِنُونَ بِمَا أُنْزِلَ إِلَيْكَ وَمَا أُنْزِلَ مِنْ قَبْلِكَ وَبِالْآخِرَةِ هُمْ يُوقِنُونَ")
        'والذين يؤمنون بما أنزل إليك وما أنزل من قبلك وبالآخرة هم يوقنون'
        >>> denoise(" ْ ً ٌ ٍ َ ُ ِ ّ ۡ ࣰ ࣱ ࣲ ٰ ")
        '              '
    """
    return re.sub(noise, "", text)


deNoise = denoise


def normalize(text, replacement_tuples=[]):
    """Normalize Arabic text by replacing complex characters by simple ones.
    The function is used internally to do batch replacements. Also, it can be called externally
     to run custom replacements with a list of tuples of (character/regex, replacement).

    Args:
        text (str): the string that needs to be normalized
        replacement_tuples (list of tuple pairs): (character/regex, replacement)

    Examples:
        >>> normalize('AlphaBet', [("A", "a"), ("B", "b")])
        'alphabet'
    """
    for pat, repl in replacement_tuples:
        text = re.sub(pat, repl, text)
    return text


def normalize_per(text):
    """Normalize Persian strings by converting Arabic chars to related Persian unicode chars.
    fixing Alifs, Alif Maqsuras, hamzas, ta marbutas, kaf, ya، Fathatan, kasra;

    Args:
        text (str): user input string to be normalized

    Examples:
        >>> normalize_per("سياسي")
        'سیاسی'
        >>> normalize_per("مدينة")
        'مدینه'
        >>> normalize_per("درِ باز")
        'در باز'
        >>> normalize_per("حتماً")
        'حتما'
        >>> normalize_per("مدرك")
        'مدرک'
        >>> normalize_per("أسماء")
        'اسما'
        >>> normalize_per("دربارۀ")
        'درباره'

    """

    repl = [
        ('ك', 'ک'),
        ('[أاإٱ]', 'ا'),
        ('[يى]ء?', 'ی'),
        ('ؤِ', 'و'),
        ('ئ', 'ی'),
        ('[ءًِ]', ''),
        ('[ۀة]', 'ه')
    ]

    return normalize(text, repl)


def normalize_ara_light(text):
    """Lighlty normalize Arabic strings:
    fixing only Alifs, Alif Maqsuras;
    replacing hamzas on carriers with standalone hamzas

    Args:
        text (str): the string that needs to be normalized

    Examples:
        >>> normalize_ara_light("ألف الف إلف آلف ٱلف")
        'الف الف الف الف الف'
        >>> normalize_ara_light("يحيى")
        'يحيي'
        >>> normalize_ara_light("مقرئ فيء")
        'مقرء فء'
        >>> normalize_ara_light("قهوة")
        'قهوة'
    """
    text = normalize_composites(text)
    repl = [("أ", "ا"), ("ٱ", "ا"), ("آ", "ا"), ("إ", "ا"),    # alifs
            ("ى", "ي"),                                        # alif maqsura
            ("يء", "ء"), ("ىء", "ء"), ("ؤ", "ء"), ("ئ", "ء"),  # hamzas
            ]
    return normalize(text, repl)
    

def normalize_ara_heavy(text):
    """Normalize Arabic text by simplifying complex characters:
    alifs, alif maqsura, hamzas, ta marbutas

    Examples:
        >>> normalize_ara_heavy("ألف الف إلف آلف ٱلف")
        'الف الف الف الف الف'
        >>> normalize_ara_heavy("يحيى")
        'يحيي'
        >>> normalize_ara_heavy("مقرئ فيء")
        'مقر في'
        >>> normalize_ara_heavy("قهوة")
        'قهوه'
    """
    text = normalize_composites(text)
    repl = [("أ", "ا"), ("ٱ", "ا"), ("آ", "ا"), ("إ", "ا"),  # alifs
            ("ى", "ي"),                                      # alif maqsura
            ("ؤ", ""), ("ئ", ""), ("ء", ""),                 # hamzas
            ("ة", "ه")                                       # ta marbuta
            ]
    return normalize(text, repl)


def normalize_composites(text, method="NFKC"):
    """Normalize composite characters and ligatures\
    using unicode normalization methods.

    Composite characters are characters that consist of
    a combination of a letter and a diacritic (e.g.,
    ؤ "U+0624 : ARABIC LETTER WAW WITH HAMZA ABOVE",
    آ "U+0622 : ARABIC LETTER ALEF WITH MADDA ABOVE").
    Some normalization methods (NFD, NFKD) decompose
    these composite characters into their constituent characters,
    others (NFC, NFKC) compose these characters
    from their constituent characters.

    Ligatures are another type of composite character:
    one unicode point represents one or more letters (e.g.,
    ﷲ "U+FDF2 : ARABIC LIGATURE ALLAH ISOLATED FORM",
    ﰋ "U+FC0B : ARABIC LIGATURE TEH WITH JEEM ISOLATED FORM").
    Such ligatures can only be decomposed (NFKC, NFKD)
    or kept as they are (NFC, NFD); there are no methods
    that compose them from their constituent characters.
    
    Finally, Unicode also contains code points for the
    different contextual forms of a letter (isolated,
    initial, medial, final), in addition to the code point
    for the general letter. E.g., for the letter ba':

    * general:	0628	ب
    * isolated:	FE8F	ﺏ
    * final:  	FE90	ﺐ
    * medial: 	FE92	ﺒ
    * initial: 	FE91	ﺑ
    
    Some methods (NFKC, NFKD)
    replace those contextual form code points by the
    equivalent general code point. The other methods
    (NFC, NFD) keep the contextual code points as they are.
    There are no methods that turn general letter code points
    into their contextual code points.

    ====== ========== ========= ================
    method composites ligatures contextual forms
    ====== ========== ========= ================
    NFC    join       keep      keep
    NFD    split      keep      keep
    NFKC   join       decompose generalize
    NFKD   split      decompose generalize
    ====== ========== ========= ================
    
    For more details about Unicode normalization methods,
    see https://unicode.org/reports/tr15/

    Args:
        text (str): the string to be normalized
        method (str): the unicode method to be used for normalization
            (see https://docs.python.org/3.5/library/unicodedata.html).
            Default: NFKC, which is most suited for Arabic. 

    Examples:
        >>> len("ﷲ") # U+FDF2: ARABIC LIGATURE ALLAH ISOLATED FORM
        1
        >>> len(normalize_composites("ﷲ"))
        4
        >>> [char for char in normalize_composites("ﷲ")]
        ['ا', 'ل', 'ل', 'ه']
        
        >>> len("ﻹ") # UFEF9: ARABIC LIGATURE LAM WITH ALEF WITH HAMZA BELOW ISOLATED FORM
        1
        >>> len(normalize_composites("ﻹ"))
        2

        alif+hamza written with 2 unicode characters:
        U+0627 (ARABIC LETTER ALEF) + U+0654 (ARABIC HAMZA ABOVE):
        
        >>> a = "أ"
        >>> len(a)
        2
        >>> len(normalize_composites(a))
        1
    """
    return unicodedata.normalize(method, text)


def denormalize(text):
    """Replace complex characters with a regex covering all variants.

    Examples:
        >>> denormalize("يحيى")
        'يحي[يى]'
        >>> denormalize("هوية")
        'هوي[هة]'
        >>> denormalize("مقرئ")
        'مقر(?:[ؤئ]|[وي]ء)'
        >>> denormalize("فيء")
        'في(?:[ؤئ]|[وي]ء)'
    """
    alifs = '[إأٱآا]'
    alif_maqsura = '[يى]\\b'
    alif_maqsura_reg = '[يى]'
    ta_marbuta = '[هة]\\b'
    ta_marbuta_reg = '[هة]'
    hamzas = '[ؤئء]'
    hamzas_reg = '(?:{}|{}ء)'.format('[ؤئ]', '[وي]')
    # Applying deNormalization
    text = re.sub(alifs, alifs, text)
    text = re.sub(alif_maqsura, alif_maqsura_reg, text)
    text = re.sub(ta_marbuta, ta_marbuta_reg, text)
    text = re.sub(hamzas, hamzas_reg, text)
    return text


#def ar_ch_len(fp):
def ar_cnt_file(fp, mode="token", incl_editor_sections=True):
    """Count the number of Arabic characters/tokens in a text, given its pth

    Args:
        fp (str): url / path to a file
        mode (str): either "char" for count of Arabic characters,
                    or "token" for count of Arabic tokens
        incl_editor_sections (bool): if False, the sections marked as editorial
            (### |EDITOR|) will be left out of the token/character count.
            Default: True (editorial sections will be counted)

    Returns:
        (int): Arabic character/token count 
    """
    splitter = "#META#Header#End#"
    try:
        with urllib.request.urlopen(fp) as f:
            book = f.read().decode('utf-8')
    except:
        with open(fp, mode="r", encoding="utf-8") as f:
            book = f.read()

    if splitter in book:
        text = book.split(splitter)[-1]
    else:
        text = book
        msg = "This text is missing the splitter!\n{}".format(fp)
        #raise Exception(msg)
    if not incl_editor_sections:
        text = re.sub(r"### \|EDITOR.+?(### |\Z)", r"\1", text,
                      flags = re.DOTALL)

    # count the number of Arabic letters or tokens:
    
    if mode == "char":
        return ar_ch_cnt(text)
    else:
        return ar_tok_cnt(text)



def ar_ch_cnt(text):
    """
    Count the number of Arabic characters in a string

    :param text: text
    :return: number of the Arabic characters in the text

    Examples:
        >>> a = "ابجد ابجد اَبًجٌدُ"
        >>> ar_ch_cnt(a)
        16
    """
    return len(ar_char.findall(text))


def ar_tok_cnt(text):
    """
    Count the number of Arabic tokens in a string

    :param text: text
    :return: number of Arabic tokens in the text

    Examples:
        >>> a = "ابجد ابجد اَبًجٌدُ"
        >>> ar_tok_cnt(a)
        3
    """
    return len(ar_tok.findall(text))


def tokenize(text, token_regex=ar_tok):
    """Tokenize a text into tokens defined by `token_regex`

    NB: make sure to remove the OpenITI header from the text

    Args:
        text (str): full text with OpenITI header removed,
            cleaned of order marks ("\u202a", "\u202b", "\u202c")
        token_regex (str): regex that defines a token
        
    Returns:
        tuple (tokens (list): list of all tokens in the text,
               tokenStarts (list): list of start index of each token
               tokenEnds (list): list of end index of each token
               )
    Examples:
        >>> a = "ابجد ابجد اَبًجٌدُ"
        >>> tokens, tokenStarts, tokenEnds = tokenize(a)
        >>> tokens
        ['ابجد', 'ابجد', 'اَبًجٌدُ']
        >>> tokenStarts
        [0, 5, 10]
        >>> tokenEnds
        [4, 9, 18]
    """
    #find matches
    tokens = [m for m in re.finditer(token_regex,text)]

    #extract tokens and start,end positions
    tokenStarts = [m.start() for m in tokens]
    tokenEnds = [m.end() for m in tokens]
    tokens = [m.group() for m in tokens]

    return tokens, tokenStarts, tokenEnds


def decode_unicode_name(s):
    """Convert unicode names into the characters they refer to.

    Args:
        s (str): input string
    
    Returns:
        str

    Examples:
        >>> decode_unicode_name("ARABIC LETTER ALEF_ARABIC LETTER YEH")
        'اي'

        >>> decode_unicode_name("ARABIC LETTER ALEF_*")
        'ا*'
    """
    new_s = ""
    for x in s.split("_"):
        try:
            new_s += unicodedata.lookup(x)
        except:
            new_s += x
    return new_s

if __name__ == "__main__":
    doctest.testmod()



"""Convert GRAR library html to OpenITI mARkdown.

This script subclasses the generic MarkdownConverter class
from the html2md module (based on python-markdownify,
https://github.com/matthewwithanm/python-markdownify),
which uses BeautifulSoup to create a flexible converter.
The subclass in this module, GRARHtmlConverter,
adds methods specifically for the conversion of books from
the GRAR library to OpenITI mARkdown:

* span conversion: the GRAR html seems to be a conversion of tei xml;
    the tei data is often embedded inside the id of a span.

Inheritance schema of the GRARHtmlConverter:

======================== ==========================
MarkdownConverter        GRARHtmlConverter
======================== ==========================
Options                  (inherited)
DefaultOptions           (inherited)
__init__                 (inherited)
__getattr__              (inherited)
convert                  (inherited)
process_tag              (inherited)
process_text             (inherited)
fill_out_columns         (inherited)
post_process_md          post_process_md (appended)
should_convert_tag       (inherited)
indent                   (inherited)
underline                (inherited)
create_underline_line    (inherited)
convert_a                (inherited)
convert_b                (inherited)
convert_blockquote       convert_blockquote
convert_br               (inherited)
convert_em               (inherited)
convert_hn               (inherited)
convert_i                (inherited)
convert_img              (inherited)
convert_list             (inherited)
convert_li               (inherited)
convert_ol               (inherited)
convert_p                (inherited)
convert_table            (inherited)
convert_tr               (inherited)
convert_ul               (inherited)
convert_strong           (inherited)
                         convert_span
======================== ==========================

"""
import re

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(path.dirname(root_folder)))
    sys.path.append(root_folder)

from openiti.new_books.convert.helper import html2md
from openiti.new_books.convert.helper.html2md import *  # import all constants!


class ThielenHtmlConverter(html2md.MarkdownConverter):
    """Convert GRAR library html to OpenITI mARkdown."""

##    def __init__(self, **options):
##        super().__init__(**options)
##        self.class_dict = dict()
##        self.class_dict["linebreak"] = '\n'

    def post_process_md(self, text):
        """Appends to the MarkdownConverter.post_process_md() method."""
        # remove blank lines marked with "DELETE_PREVIOUS_BLANKLINES" tag
        text = re.sub("\n*(PageV\d+P\d+) *@QUOTE@", r"\n\n\1\n\n# ", text)
        text = re.sub("[\r\n]@QUOTE@", "\n#", text)
        text = re.sub("@QUOTE@", "", text)
        text = re.sub(" *\n~~ *\n", "\n", text)
        text = super().post_process_md(text)
        return text

    def convert_blockquote(self, el, text):
        """Convert blockquote tags to mARkdown

        NB: the @QUOTE@ tag is a temporary tag that will be removed
        in the post-processing step.
        
        Examples:
            >>> import html2md_GRAR
            >>> h = 'abc <blockquote>def</blockquote> ghi'
            >>> html2md_GRAR.markdownify(h)
            'abc def ghi'

            >>> h = 'abc <span id="pb-21"/><blockquote>def</blockquote> ghi jkl'
            >>> html2md_GRAR.markdownify(h)
            'abc\\n\\nPageV00P020\\n\\n# def ghi jkl'


        """
        return "@QUOTE@"+text

    def convert_page(self, el, text):
        """Convert page tags"""
        return " "+text+" "

    def convert_head(self, el, text):
        """Convert <head> tags"""
        if "rendition" in el.attrs:
            try:
                n = int(el["rendition"][-1])
            except:
                n = 1
            return "\n### {} {}\n".format("|" * n, text or "")
        else:
            return "\n### ||| "

    def convert_lb(self, el, text):
        return "\n~~"
    
    def convert_span(self, el, text):
        """Converts html <span> tags, depending on their id attribute.

        Example:
            >>> import html2md_GRAR
            >>> h = 'abc <span>def</span> ghi'
            >>> html2md_GRAR.markdownify(h)
            'abc def ghi'

            >>> h = 'abc <span class="unknown_span_class">def</span> ghi'
            >>> html2md_GRAR.markdownify(h)
            'abc def ghi'

            Page numbers (NB: mARkdown uses page end instead of page beginning)

            >>> h = 'abc <span id="pb-21"/>def  ghi jkl'
            >>> html2md_GRAR.markdownify(h)
            'abc PageV00P020 def ghi jkl'

            Sections: 

            >>> h = 'abc <span class="book" id="part-2 div1-2"/>def  ghi jkl'
            >>> html2md_GRAR.markdownify(h)
            'abc\\n\\n### | [book 2]\\n\\ndef ghi jkl'

            >>> h = 'abc <span class="chapter" id="part-2 div2-1"/>def  ghi jkl'
            >>> html2md_GRAR.markdownify(h)
            'abc\\n\\n### || [chapter 1]\\n\\ndef ghi jkl'

            >>> h = 'abc <span class="chapter" id="part-2 div2-1" title="Intro"/>def  ghi jkl'
            >>> html2md_GRAR.markdownify(h)
            'abc\\n\\n### || [chapter 1: Intro]\\n\\ndef ghi jkl'

        """
        try:
            if "pb" in el["id"]:
                #print(el["id"])
                page_no = re.findall("\d+", el["id"])[-1]
                #print(page_no)
                page_no = int(page_no) - 1
                return "PageV00P{:03d} ".format(page_no)
            elif "div"  in el["id"]:
                no = int(re.findall("div(1|2)", el["id"])[0])
                #print("div1 or div2?", no)
                div_no = re.findall("\d+", el["id"])[-1]
                #print("div no:", div_no)
                try:
                    return "\n\n### {} [{} {}: {}]\n\n".format(\
                        "|"*no, el["class"][0], div_no, el["title"])
                except:
                    #print("\n\n### {} [{} {}]\n\n".format(\
                    #    "|"*no, el["class"][0], div_no))
                    return "\n\n### {} [{} {}]\n\n".format(\
                        "|"*no, el["class"][0], div_no)
        except Exception as e:
            #print(e)
            #input()
            pass
        try:
            if el["title"] == "linebreak":
                return "\n~~"
        except:
            pass
        return text


def markdownify(html, **options):
    """Shortcut to the convert method of the HindawiConverter class."""
    return ThielenHtmlConverter(**options).convert(html)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""Convert Hindawi library html to OpenITI mARkdown.

This script subclasses the generic MarkdownConverter class
from the html2md module (based on python-markdownify,
https://github.com/matthewwithanm/python-markdownify),
which uses BeautifulSoup to create a flexible converter.
The subclass in this module, HindawiConverter,
adds methods specifically for the conversion of books from
the Hindawi library to OpenITI mARkdown:
* special treatment of <h4> heading tags
* div classes "poetry_container", "section", "footnote" and "subtitle"
* span class "quran"

The easiest way to use this is to simply feed the html (as string)
to the markdownify() function, which will create an instance
of the HindawiConverter class and return the converted string.

Examples (doctests):

    Headings: h1 (from superclass html2md.MarkdownConverter)

    >>> import html2md_hindawi
    >>> h = '<h1>abc</h1>'
    >>> html2md_hindawi.markdownify(h)
    '\\n\\n### | abc\\n\\n'

    NB: heading style is OpenITI mARkdown style by default,
        but can be set to other styles as well:

    >>> h = '<h1>abc</h1>'
    >>> html2md_hindawi.markdownify(h, md_style=UNDERLINED)
    '\\n\\nabc\\n===\\n\\n'

    >>> h = '<h1>abc</h1>'
    >>> html2md_hindawi.markdownify(h, md_style=ATX)
    '\\n\\n# abc\\n\\n'

    Headings: <h4>
    
    NB: in the Hindawi library, <h4> tag is used for section headings
        on all levels but the highest one.
        The section level must be derived from the id of the parent div. 

    >>> h = '<div class="section" id="sect2_4"><h4>abc</h4></div>'
    >>> html2md_hindawi.markdownify(h)
    '\\n\\n### ||| abc\\n\\n'

    >>> h = '<div class="section" id="sect5_2"><h4>abc</h4></div>'
    >>> html2md_hindawi.markdownify(h)
    '\\n\\n### |||||| abc\\n\\n'


    Poetry div, single-line:
    
    >>> h = '\
    <div class="poetry_container line">\
      <div>\
        <div>hemistich1</div>\
        <div>hemistich2</div>\
      </div>\
    </div>'
    >>> html2md_hindawi.markdownify(h)
    '\\n# hemistich1 %~% hemistich2\\n'
    
    Poetry div, multiple line:

    >>> h = '\
    abc\
    <div class="poetry_container">\
      <div>\
        <div>hemistich1</div>\
        <div>hemistich2</div>\
      </div>\
      <div>\
        <div>hemistich3</div>\
        <div>hemistich4</div>\
      </div>\
    </div>\
    def'
    >>> html2md_hindawi.markdownify(h)
    ' abc\\n# hemistich1 %~% hemistich2\\n# hemistich3 %~% hemistich4\\ndef'
    
    Section div without heading:
    
    >>> h = 'abc\
             <div class="section" id="sect2_9">def</div>\
             ghi'
    >>> html2md_hindawi.markdownify(h)
    'abc\\n\\n### |||\\ndef\\n\\nghi'

    Section div with heading (handled by h4):
    
    >>> h = 'abc\
             <div class="section" id="sect2_9">\
               <h4>title</h4>\
               <p>def</p>\
             </div>\
             ghi'
    >>> html2md_hindawi.markdownify(h)
    'abc\\n\\n### ||| title\\n\\n# def\\n\\nghi'

    Footnote divs:
    
    >>> h = '<div class="footnote"><sup>1 </sup>footnotetext</div>'
    >>> html2md_hindawi.markdownify(h)
    '\\n\\nFOOTNOTE1 footnotetext\\n\\n'

    NB: FOOTNOTE is a tag that will be used to extract all footnotes
        in a next step.

    Subtitle divs:

    >>> h = '<h1>Title text</h1><div class="subtitle">Subtitle text</div>'
    >>> html2md_hindawi.markdownify(h)
    '\\n\\n### | Title text Subtitle text\\n\\n'

    Divs without class or with an unsupported class are simply stripped:
    
    >>> h = 'abc\
             <div>def</div>\
             ghi'
    >>> html2md_hindawi.markdownify(h)
    'abc def ghi'
        
    >>> h = 'abc\
             <div class="unknown_div_class">def</div>\
             ghi'
    >>> html2md_hindawi.markdownify(h)
    'abc def ghi'

    Spans with class "quran":

    >>> h = 'abc <span class="quran">def ghi</span> jkl'
    >>> html2md_hindawi.markdownify(h)
    'abc @QUR02 def ghi jkl'

    # the latter is a result of post-processing;
    # the function itself will produce: 
    # 'abc @QUR@ def ghi\\njkl'

    Spans without class or with an unsupported class are stripped:
    
    >>> h = 'abc <span>def</span> ghi'
    >>> html2md_hindawi.markdownify(h)
    'abc def ghi'
    
    >>> h = 'abc <span class="unknown_span_class">def</span> ghi'
    >>> html2md_hindawi.markdownify(h)
    'abc def ghi'


    Links: 

    >>> h = '<a href="a/b/c">abc</a>'
    >>> html2md_hindawi.markdownify(h)
    '[abc](a/b/c)'

    
    Unordered lists: 

    >>> h = '<ul><li>item1</li><li>item2</li></ul>'
    >>> html2md_hindawi.markdownify(h)
    '\\n* item1\\n* item2\\n\\n'
    
    Ordered lists:

    >>> h = '<ol><li>item1</li><li>item2</li></ol>'
    >>> html2md_hindawi.markdownify(h)
    '\\n1. item1\\n2. item2\\n\\n'

    Nested lists:
    
    >>> h = '<ol><li>item1</li><li>item2:<ul><li>item3</li><li>item4</li></ul></li></ol>'
    >>> html2md_hindawi.markdownify(h)
    '\\n1. item1\\n2. item2:\\n\\n\\t* item3\\n\\t* item4\\n\\t\\n\\n'


    Italics (<i> and <em> tags):

    >>> h = 'abc <em>def</em> ghi'
    >>> html2md_hindawi.markdownify(h)
    'abc *def* ghi'

    >>> h = 'abc <i>def</i> ghi'
    >>> html2md_hindawi.markdownify(h)
    'abc *def* ghi'


    Bold (<b> and <strong> tags):

    >>> h = 'abc <b>def</b> ghi'
    >>> html2md_hindawi.markdownify(h)
    'abc **def** ghi'

    >>> h = 'abc <strong>def</strong> ghi'
    >>> html2md_hindawi.markdownify(h)
    'abc **def** ghi'


    Tables:
    
    >>> h = '\
    <table>\
      <tr>\
        <th>th1aaa</th><th>th2</th>\
      </tr>\
      <tr>\
        <td>td1</td><td>td2</td>\
      </tr>\
    </table>'
    >>> html2md_hindawi.markdownify(h)
    '\\n\\n| th1aaa | th2 |\\n| ------ | --- |\\n| td1    | td2 |\\n\\n'

"""
import re


if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(path.dirname(root_folder)))
    sys.path.append(root_folder)

from openiti.new_books.convert.helper import html2md
from openiti.new_books.convert.helper.html2md import *  # import all constants!


class HindawiConverter(html2md.MarkdownConverter):
    """Convert Hindawi library html to OpenITI mARkdown."""

    def __init__(self, **options):
        super().__init__(**options)
        self.class_dict = dict()
        self.class_dict["quran"] = '@QUR@ {text}\n'
        self.class_dict["subtitle"] = 'DELETE_PREVIOUS_BLANKLINES {text}\n\n'
        self.class_dict["footnote"] = '\n\nFOOTNOTE{text}\n\n'

##    def process_Quran(self, match):
##        """Reformat Quran quote matches to mARkdown named entity standard."""
##        
##        prec_char = match.group(0)[0]
##        if prec_char == "\n":
##            prec_char = "\n\n# "
##        foll_char = match.group(0)[-1]
##        quote = match.group(0)[7:-2]
##        quote_words = len(re.findall(" +|\n", quote)) + 1
##        #print(match.group(0))
##        #print(quote)
##        return "{}@QUR0{} {}{}".format(prec_char, quote_words, quote, foll_char)
##
##
##    def post_process_md(self, text):
##        """Appends to the MarkdownConverter.post_process_md() method."""
##        # remove blank lines marked with "DELETE_PREVIOUS_BLANKLINES" tag
##        text = re.sub(r"\n+DELETE_PREVIOUS_BLANKLINES", "", text)
##        text = re.sub(".@QUR@ .+?\n[^~]", self.process_Quran, text,
##                      flags=re.DOTALL)
##        text = super().post_process_md(text)
##        return text

    def get_section_level(self, el):
        """Gets the level of the current section (or its parent)."""
        sect_level = 8
        try:
            if "section" in el["class"]:
                sectID = el["id"]
                sect_level, sect_no = re.findall("\d+", sectID)
                sect_level = int(sect_level)
                #print(sect_level)
            else:
                try:
                    sect_level = self.get_section_level(el.parent)
                except:
                    pass
        except:
                try:
                    sect_level = self.get_section_level(el.parent)
                except:
                    pass
        return sect_level

    #------------------------------------------------------------------------
    # new conversion methods 
    # (and adaptations of conversion methods from the super-class)

    def convert_a(self, el, text):
        """Converts html links.

        Overwrites the MarkdownConverter.post_process_md() method.
        Introduces an exception for links between footnote markers
        and footnootes. 

        Example:
            >>> import html2md_hindawi
            >>> h = '<a href="a/b/c">abc</a>'
            >>> html2md_hindawi.markdownify(h)
            '[abc](a/b/c)'

            >>> import html2md_hindawi
            >>> h = 'abc <a href="ftn1">1</a>'
            >>> html2md_hindawi.markdownify(h)
            'abc [1]'
        """
        href = el.get('href')
        if "ftn" in href or "fn" in href:
            # do not include a link for footnotes:
            return '[%s]' % text
        else:
            # return a markdown representation of a link: [text](href)
            return super().convert_a(el, text)

    def convert_div(self, el, text):
        """Converts html <div> tags, depending on their class.

        In the MarkdownConverter class, div tags are simply stripped away.
        
        Examples:
        
            # no div class: tags are stripped off
            
            >>> import html2md_hindawi
            >>> h = 'abc\
                     <div>def</div>\
                     ghi'
            >>> html2md_hindawi.markdownify(h)
            'abc def ghi'
            
            # unknown div class: tags are stripped off
            
            >>> import html2md_hindawi
            >>> h = 'abc\
                     <div class="unknown_div_class">def</div>\
                     ghi'
            >>> html2md_hindawi.markdownify(h)
            'abc def ghi'
            
            # poetry single-line:
            
            >>> h = '\
            <div class="poetry_container line">\
              <div>\
                <div>hemistich1</div>\
                <div>hemistich2</div>\
              </div>\
            </div>'
            >>> html2md_hindawi.markdownify(h)
            '\\n# hemistich1 %~% hemistich2\\n'
            
            # poetry multiple line:

            >>> h = '\
            abc\
            <div class="poetry_container">\
              <div>\
                <div>hemistich1</div>\
                <div>hemistich2</div>\
              </div>\
              <div>\
                <div>hemistich3</div>\
                <div>hemistich4</div>\
              </div>\
            </div>\
            def'
            >>> html2md_hindawi.markdownify(h)
            ' abc\\n# hemistich1 %~% hemistich2\\n# hemistich3 %~% hemistich4\\ndef'
            
            # section without heading:
            
            >>> h = 'abc\
                     <div class="section" id="sect2_9">def</div>\
                     ghi'
            >>> html2md_hindawi.markdownify(h)
            'abc\\n\\n### |||\\ndef\\n\\nghi'

            # section with heading (handled by h4):
            
            >>> h = 'abc\
                     <div class="section" id="sect2_9">\
                       <h4>title</h4>\
                       <p>def</p>\
                     </div>\
                     ghi'
            >>> html2md_hindawi.markdownify(h)
            'abc\\n\\n### ||| title\\n\\n# def\\n\\nghi'

            # footnote:
            
            >>> h = '<div class="footnote"><sup>1 </sup>footnotetext</div>'
            >>> html2md_hindawi.markdownify(h)
            '\\n\\nFOOTNOTE1 footnotetext\\n\\n'

            Paragraph block (similar to <p>):

            >>> h = '<div class="paragraph-block">abc def ghi</div>'
            >>> html2md_hindawi.markdownify(h)
            '\\n\\n# abc def ghi\\n\\n'

            >>> h = '<div class="paragraph-block"><p>abc def ghi</p></div>'
            >>> html2md_hindawi.markdownify(h)
            '\\n\\n# abc def ghi\\n\\n'

            >>> h = '<div class="paragraph-block"></div>'
            >>> html2md_hindawi.markdownify(h)
            ''

            >>> h = '<div class="paragraph-block"><div class="paragraph-block">abc</div></div>'
            >>> html2md_hindawi.markdownify(h)
            '\\n\\n# abc\\n\\n'

        """
        try:
            el["class"]
        except:
            return text
        if "poetry_container" in el["class"]:
            t = []
            for div in el.find_all("div", recursive=False):
                div_list = [x.text for x in div.find_all("div", recursive=False)]
                t.append(" %~% ".join(div_list))
            return "\n# {}\n".format("\n# ".join(t))
        elif "section" in el["class"]:
            # deal with untitled sections: 
            if not el.find(re.compile("h\d")):
                section_level = self.get_section_level(el)+1
                pipes = "|"*section_level
                return '\n\n### %s \n%s\n\n' % (pipes, text)
            #else: handled by convert_hn
        elif "paragraph-block" in el["class"]:
            if not el.find("p"):
                if not el.find("div"):
                    return '\n\n# %s\n\n' % text if text else ''
        else:
            for c in el["class"]:
                if c in self.class_dict:
                    return self.class_dict[c].format(text=text, el=el, c=c) 
           
        return text

    def convert_hn(self, n, el, text):
        """Converts headings in the usual way except h4 headings.

        In the Hindawi library, <h4> tags
        are used for subsections on any level.
        The section level must be derived from the id of the parent div.
        
        Example:
            >>> import html2md_hindawi
            >>> h = '<h1>abc</h1>'
            >>> html2md_hindawi.markdownify(h)
            '\\n\\n### | abc\\n\\n'

            >>> h = '<h3>abc</h3>'
            >>> html2md_hindawi.markdownify(h)
            '\\n\\n### ||| abc\\n\\n'

            >>> h = '<div class="section" id="sect5_2"><h4>abc</h4></div>'
            >>> html2md_hindawi.markdownify(h)
            '\\n\\n### |||||| abc\\n\\n'
        """

        if n == 4:
            return self.convert_h4(el, text)
        else:
            return super().convert_hn(n, el, text)
            
    def convert_h4(self, el, text):
        """Converts <h4> header tags.

        In the Hindawi library, <h4> tags are used
        for subsections on any level.
        The section level must be taken from the id of the parent div.
        NB: the ### | level in Hindawi is 0.

        Example:
            >>> import html2md_hindawi
            >>> h = '<div class="section" id="sect2_4"><h4>abc</h4></div>'
            >>> html2md_hindawi.markdownify(h)
            '\\n\\n### ||| abc\\n\\n'

            >>> h = '<div class="section" id="sect5_2"><h4>abc</h4></div>'
            >>> html2md_hindawi.markdownify(h)
            '\\n\\n### |||||| abc\\n\\n'
        """
        
        section_level = self.get_section_level(el)
        hashes = "#"*(section_level+1)
        style = self.options['md_style']
        if style == ATX_CLOSED:
            return '\n\n#%s %s %s\n\n' % (hashes, text, hashes)
        elif style == ATX:
            return '\n\n%s %s\n\n' % (hashes, text)
        elif style == OPENITI:
            return '\n\n### %s %s\n\n' % ("|"*(section_level+1), text)
        return '\n\n%s %s\n\n' % (hashes, text)

    def convert_span(self, el, text):
        """Converts html <span> tags, depending on their class attribute.

        Supported span classes should be stored in self.span_dict
        (key: span class (str); value: formatting string)

        Example:
            >>> import html2md_hindawi
            >>> h = 'abc <span>def</span> ghi'
            >>> html2md_hindawi.markdownify(h)
            'abc def ghi'
            
            >>> h = 'abc <span class="unknown_span_class">def</span> ghi'
            >>> html2md_hindawi.markdownify(h)
            'abc def ghi'
            
            >>> h = 'abc <span class="quran">def  ghi</span> jkl'
            >>> html2md_hindawi.markdownify(h)
            'abc @QUR02 def ghi jkl'

            # the latter is a result of post-processing;
            # the function itself will produce: 
            # 'abc @QUR@ def ghi\\njkl'
        """
        try:
            for c in el["class"]:
                if c in self.class_dict:
                    if text:
                        return self.class_dict[c].format(text=text, el=el, c=c)
                    else:
                        return ''
        except Exception as e:
            #print(e)
            pass
        return text



def markdownify(html, **options):
    """Shortcut to the convert method of the HindawiConverter class."""
    return HindawiConverter(**options).convert(html)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



### subclassing experiment: 
##
##import re
##
##class Html2md(object):
##    def post_process(self, text):
##        # remove leading and trailing spaces in lines:
##        text = re.sub(r" *(\n+) *", r"\1", text)
##        # remove unwanted additional spaces and lines:
##        text = re.sub(r"\n{3,}", r"\n\n", text)
##        text = re.sub(r" +", r" ", text)
##        return text
##
##class html2md_hindawi(Html2md):
##    def post_process(self, text):
##        text = super().post_process(text)
##        # remove blank lines marked with "DELETE_PREVIOUS_BLANKLINES" tag
##        text = re.sub(r"\n+DELETE_PREVIOUS_BLANKLINES", "", text)
##        # replace placeholders for spaces in tables: 
##        text = re.sub("ç", " ", text)
##        return text    
##
##    def post_process_super(self, text):
##        text = super().post_process(text)
##        return text
##
##
##test = """ab \ncd\n ef\n\n\n\ngh          ij\n\nDELETE_PREVIOUS_BLANKLINESklçççmn"""
##h1 = Html2md()
##
##h2 = html2md_hindawi()
##print(h1.post_process(test))
##print("+"*40)
##print(h2.post_process(test))
##print("+"*40)
##print(h2.post_process_super(test))







"""Convert GRAR library html to OpenITI mARkdown.

This script subclasses the generic MarkdownConverter class
from the html2md module (based on python-markdownify,
https://github.com/matthewwithanm/python-markdownify),
which uses BeautifulSoup to create a flexible converter.
The subclass in this module, PtolHtmlConverter,
adds methods specifically for the conversion of books from
the GRAR library to OpenITI mARkdown:

* span conversion: the GRAR html seems to be a conversion of tei xml;
    the tei data is often embedded inside the id of a span.

Inheritance schema of the PtolHtmlConverter:

======================== ==========================
MarkdownConverter        PtolHtmlConverter
======================== ==========================
Options                  (inherited)
DefaultOptions           (inherited)
__init__                 (inherited)
__getattr__              (inherited)
convert                  (inherited)
process_tag              (inherited)
process_text             (inherited)
fill_out_columns         (inherited)
post_process_md          (inherited)
should_convert_tag       (inherited)
indent                   (inherited)
underline                (inherited)
create_underline_line    (inherited)
convert_a                convert_a
convert_b                (inherited)
convert_blockquote       (inherited)
convert_br               (inherited)
convert_em               (inherited)
convert_hn               (inherited)
convert_i                (inherited)
convert_img              (inherited)
convert_list             (inherited)
convert_li               (inherited)
convert_ol               (inherited)
convert_p                (inherited)
convert_table            (inherited)
convert_tr               (inherited)
convert_ul               (inherited)
convert_strong           (inherited)
                         convert_editor
                         convert_small
                         convert_span
                         convert_u
======================== ==========================

"""
import re

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(path.dirname(root_folder)))
    sys.path.append(root_folder)

from openiti.new_books.convert.helper import html2md
from openiti.new_books.convert.helper.html2md import *  # import all constants!


class PtolHtmlConverter(html2md.MarkdownConverter):
    """Convert GRAR library html to OpenITI mARkdown."""

##    def __init__(self, **options):
##        super().__init__(**options)
##        self.class_dict = dict()
##        self.class_dict["linebreak"] = '\n'


    def convert_a(self, el, text):
        """Convert html links to markdown-style links.

        Example:
            >>> import html2md
            >>> h = '<a href="a/b/c">abc</a>'
            >>> html2md.markdownify(h)
            '[abc](a/b/c)'
        """
        class_ = el.attrs.get('class', None) or ''
        href = el.attrs.get('href', None) or ''
        title = el.attrs.get('title', None) or ''
        if "pagina" in class_:
            p, side = re.findall("(\d+)([vr]*)", el["id"])[0]
            print(el["id"], p, side)
            return "PageBegV00P{0:03d}{1}".format(int(p), side)
        if href.startswith(("#", "intus")):
            return text
        if self.options['autolinks'] and text == href and not title:
            # Shortcut syntax
            return '<%s>' % href
        title_part = ' "%s"' % title.replace('"', r'\"') if title else ''
        return '[%s](%s%s)' % (text or '', href, title_part) if href else text or ''

    def convert_editor(self, el, text):
        return "### |EDITOR|\n%s\n### | \n\n" % text

    def convert_small(self, el, text):
        class_ =  el.get("class")
        if "note" in class_:
            return " (¬{}) ".format(el["data-sign"])
        else:
            return ""

    def convert_span(self, el, text):
        """Converts html <span> tags, depending on their id attribute.

        Example:
            >>> import html2md_GRAR
            >>> h = 'abc <span>def</span> ghi'
            >>> html2md_GRAR.markdownify(h)
            'abc def ghi'

            >>> h = 'abc <span class="unknown_span_class">def</span> ghi'
            >>> html2md_GRAR.markdownify(h)
            'abc def ghi'

            Page numbers (NB: mARkdown uses page end instead of page beginning)

            >>> h = 'abc <span id="pb-21"/>def  ghi jkl'
            >>> html2md_GRAR.markdownify(h)
            'abc PageV00P020 def ghi jkl'

            Sections: 

            >>> h = 'abc <span class="book" id="part-2 div1-2"/>def  ghi jkl'
            >>> html2md_GRAR.markdownify(h)
            'abc\\n\\n### | [book 2]\\n\\ndef ghi jkl'

            >>> h = 'abc <span class="chapter" id="part-2 div2-1"/>def  ghi jkl'
            >>> html2md_GRAR.markdownify(h)
            'abc\\n\\n### || [chapter 1]\\n\\ndef ghi jkl'

            >>> h = 'abc <span class="chapter" id="part-2 div2-1" title="Intro"/>def  ghi jkl'
            >>> html2md_GRAR.markdownify(h)
            'abc\\n\\n### || [chapter 1: Intro]\\n\\ndef ghi jkl'

        """
        class_ = el.attrs.get('class', None) or ''
        id_ = el.attrs.get('id', None) or ''
        if class_ == "note":
            return ""
        elif "pagina" in class_:
            p, side = re.findall("(\d+)([vr]*)", id_)[0]
            print(id_, p, side)
            return "PageBegV00P{0:03d}{1}".format(int(p), side)
        return text

    def convert_u(self, el, text):
        #return ' @NUM@ %s\n' % text.strip() if text else ''
        return ' _NUM_ %s' % text.strip() if text else ''



def markdownify(html, **options):
    """Shortcut to the convert method of the HindawiConverter class."""
    return PtolHtmlConverter(**options).convert(html)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""A generic converter for texts into OpenITI format.

This generic converter contains the basic procedure for
converting texts into OpenITI format.

These are the main methods of the GenericConverter class:

* `convert_file(source_fp)`: basic procedure for converting an html file.
* `convert_files_in_folder(source_folder)`: convert all files files in the
    `source_folder` (calls `convert_file`)

The procedure can be used as a template for other converters.
Subclass the GenericConverter to make a converter for a new
source of scraped books::

    class GenericHtmlConverter(GenericConverter):
        def new_function(self, text):
            "This function adds functionality to GenericConverter"

Usually, it suffices to overwrite a small number of methods
of the GenericConverter class in the sub-class, and/or add
a number of methods needed for the conversion of a specific
group of documents: e.g.:: 

    class GenericHtmlConverter(GenericConverter):
        def get_data(self, source_fp):
            # code written here overwrites the GenericConverter's get_data method

        def post_process(self, text):
            # append new actions by appending them to
            # the superclass's post_process method:
            text = super().post_process(text)
            text = re.sub("\\n\\n+", "\\n\\n", text)
            return text

        def new_function(self, text):
            "This function adds functionality to GenericConverter"

The GenericConverter contains a number of dummy (placeholder) functions
that must be overwritten in the subclass because they are dependent on
the data format.

For example, this is the inheritance schema of the EShiaHtmlConverter,
which is a subclass of the GenericHtmlConverter:
(methods of GenericConverter are inherited by GenericHtmlConverter;
and methods of GenericHtmlConverter are inherited by EShiaHtmlConverter.
Methods of the superclass with the same name
in the subclass are overwritten by the latter)

================================== ========================== ==========================
GenericConverter                   GenericHtmlConverter       EShiaHtmlConverter
================================== ========================== ==========================
__init__                           __init__                   (inherited)
convert_files_in_folder            (inherited)                (inherited)
convert file                       (inherited)                (inherited)
make_dest_fp                       (inherited - generic!)     (inherited - generic!)
get_metadata (dummy)               (inherited - dummy!)       get_metadata
get_data                           get_data                   (inherited)
pre_process                        (inherited)                (inherited)
add_page_numbers (dummy)           (inherited - dummy!)       add_page_numbers
add_structural_annotations (dummy) add_structural_annotations add_structural_annotations
remove_notes (dummy)               remove_notes               remove_notes
reflow                             (inherited)                (inherited)
add_milestones (dummy)             (inherited - dummy!)       (inherited - dummy!)
post_process                       (inherited - generic!)     post_process
compose                            (inherited)                (inherited)
save_file                          (inherited)                (inherited)
                                   inspect_tags_in_html       (inherited)
                                   inspect_tags_in_folder     (inherited)
                                   find_example_of_tag        (inherited)
================================== ========================== ==========================

Examples:
    >>> from generic_converter import GenericConverter
    >>> conv = GenericConverter()
    >>> conv.VERBOSE = False
    >>> folder = r"test/"
    >>> conv.convert_file(folder+"86596.html")
    >>> conv.convert_files_in_folder(folder, ["html"])

"""

import os
import re
import textwrap

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(root_folder))
    sys.path.append(root_folder)

from openiti.helper.ara import deNoise, normalize, normalize_composites

class GenericConverter(object):

    def __init__(self, dest_folder=None, overwrite=True):
        """Initialize the default values."""
        self.magic_value = "######OpenITI#\n\n\n"
        self.header_splitter = "\n\n#META#Header#End#\n\n"
        self.endnote_splitter = "\n\n### |EDITOR|\nENDNOTES:\n\n"
        #self.dest_folder = "converted"
        if dest_folder == None:
            self.dest_folder = "converted"
        else:
            self.dest_folder = dest_folder
        self.extension = ".automARkdown"
        
        # settings for line wrapping: max number of characters in a line

        self.max_line_len = 72

        # settings for the chunking of the text:

        self.chunk_length = 300
        self.milestone_tag = "Milestone300"

        # regexes describing Arabic characters and tokens:

        self.ara_char = "[؀-ۿࢠ-ࣿﭐ-﷿ﹰ-﻿]"
        ara_tok = "{}+|[^{}+".format(self.ara_char, self.ara_char[1:])
        self.ara_tok = re.compile(ara_tok)

        self.VERBOSE = False
        self.metadata_file = "inline"
        self.OVERWRITE = overwrite

##        self.conversion_procedure(source_fp, dest_dir)


    def filter_files_in_folder(self, source_folder, extensions=[],
                               exclude_extensions=[], fn_regex=None):
        """Make a list of all the files that have (or don't have) \
        specific extensions, and/or match a specific regex pattern.

        Args:
            source_folder (str): path to the folder.
            extensions (list): if defined, only files with one of the \
                extensions in the list will be used.
                Defaults to an empty list
                (=> all files in the `source_folder` will be used)
            exclude_extensions (list): if defined
                (and no `extensions` list is defined),
                only files without one of the extensions in this list
                will be usedwill.
                Defaults to an empty list
                (=> all files in the `source_folder` will be used)
            fn_regex (str): regular expression defining the filename pattern
                e.g., "-(ara|per)\d". If `fn_regex` is defined,
                only files whose filename matches the pattern will be converted.
                Defaults to None
                (=> all files in the `source_folder` will be used)

        Returns:
            fp_list (list): a list of file paths that comply with the user's parameters
        """
        # make sure all extensions (except "") start with a period:
        extensions = [re.sub("^\.?(\w+)", r".\1", ext) for ext in extensions]
        exclude_extensions = [re.sub("^\.?(\w+)", r".\1", ext) \
                              for ext in exclude_extensions]
        
        fp_list = []
        for fn in os.listdir(source_folder):
            fp = os.path.join(source_folder, fn)
            if os.path.isfile(fp):
                incl = True
                ext = os.path.splitext(fn)[-1]
                if fn_regex:
                    if not re.findall(fn_regex, fn):
                        incl = False
                if extensions:
                    if ext not in extensions:
                        incl = False
                if exclude_extensions:
                    if ext in exclude_extensions:
                        incl = False
                if incl:
                    fp_list.append(fp)
        return fp_list


    def convert_files_in_folder(self, source_folder, dest_folder=None,
                                extensions=[], exclude_extensions=[],
                                fn_regex=None):
        """Convert all files in a folder to OpenITI format.\
        Use the `extensions` and `exclude_extensions` lists to filter\
        the files to be converted.

        Args:
            source_folder (str): path to the folder that contains
                the files that must be converted.
            extensions (list): list of extensions; if this list is not empty,
                only files with an extension in the list should be converted.
            exclude_extensions (list): list of extensions;
                if this list is not empty,
                only files whose extension is not in the list will be converted.
            fn_regex (str): regular expression defining the filename pattern
                e.g., "-(ara|per)\d". If `fn_regex` is defined,
                only files whose filename matches the pattern will be converted.

        Returns:
            None
        """
        if dest_folder:
            self.dest_folder = dest_folder
        fp_list = self.filter_files_in_folder(source_folder, extensions,
                                              exclude_extensions, fn_regex)
        for fp in fp_list:
            self.convert_file(fp)


    def convert_file(self, source_fp, dest_fp=None):
        """Convert one file to OpenITI format.

        Args:
            source_fp (str): path to the file that must be converted.

        Returns:
            None
        """
        if self.VERBOSE:
            print("converting", source_fp)

        # make the file path for the converted file
        # (based on the source_fp and the self.dest_folder value;
        # if no specific self.dest_folder is given, a `converted` folder
        # will be created in the folder of the source file)
        if not dest_fp:
            dest_fp = self.make_dest_fp(source_fp)
        self.source_fp = source_fp

        if not self.OVERWRITE and os.path.exists(dest_fp):
            print("Not overwriting already existing file:", dest_fp)
            print("Set self.overwrite to True if you want to overwrite")
            return

##        # get the metadata from the source_fp and format it
##        # WARNING: this is a dummy method, returning
##        #          a string with magic_value and header_splitter only.
##        #          Overwrite this method in subclass!
##        self.metadata = self.get_metadata(source_fp)

        # Get the raw data from the source_fp
        # WARNING: this method only reads the file at source_fp
        #          and returns it as a string.
        #          Overwrite this method in subclass if necessary!
        text = self.get_data(source_fp)

        # get the metadata from the text or from a separate file,
        # and format it.
        # WARNING: this is a dummy method, returning
        #          a string with magic_value and header_splitter only.
        #          Overwrite this method in subclass!
        if self.metadata_file == "inline":
            self.metadata = self.get_metadata(text)
        else:
            self.metadata = self.get_metadata(self.metadata_file)

        # Remove unwanted features from the text:
        # NB: this method executes a standard cleaning of the text.
        #     If the text needs more pre-processing,
        #     append additional operations to the method in the sub-class.
        text = self.pre_process(text)

        # add page numbers to the text, or convert them to OpenITI format:
        # WARNING: this is a dummy function, and returns the text without
        #          making any changes. Overwrite this method in the sub-class
        #          if page numbers need to be added to the text.
        text = self.add_page_numbers(text, source_fp)

##        # add structural tags in OpenITI format to the text:
##        # WARNING: this is a dummy function, and returns the text without
##        #          making any changes. Overwrite this method in the sub-class
##        #          if structural annotations need to be added to the text.
##        text = self.add_structural_annotations(text)

        # remove the footnotes from the text, and convert them into endnotes
        # WARNING: this is a dummy function, and returns the text without
        #          making any changes, and an empty string for the endnotes.
        #          Overwrite this method in the sub-class if notes need to be
        #          extracted from the text.
        text, notes = self.remove_notes(text)

        # add structural tags in OpenITI format to the text:
        # WARNING: this is a dummy function, and returns the text without
        #          making any changes. Overwrite this method in the sub-class
        #          if structural annotations need to be added to the text.
        text = self.add_structural_annotations(text)

        # wrap long lines (i.e., break long lines down)
        # and mark the start of paragraphs with "# ":
        # NB: this function should not be overwritten in the sub-class.
        text = self.reflow(text)

        # Add milestones to the text.
        # NB: the add_milestones method is old and needs to be replaced
        #     with the most recent function written by Masoumeh!
        #     It has been disabled and now returns the text without changes.
        text = self.add_milestones(text)

        # Make last changes to the text;
        # NB: this default method only takes care of paragraph marks.
        #     Write a post_processing method for your sub_class if needed.
        text = self.post_process(text)

        # Join the formatted metadata, text and notes strings:
        text = self.compose(self.metadata, text, notes)

        # Save the file to the dest_fp:
        self.save_file(text, dest_fp)


    def make_dest_fp(self, source_fp):
        """Make a filepath for the converted text file, \
        based on the `source_fp` and the class's `self.dest_folder`.\
        If no specific self.dest_folder is given,
        a `converted` folder is created in the source folder.

        Args:
            source_fp (str): path to the file that must be converted.

        Returns:
            dest_fp (str): path to the converted text file.
        """
        source_folder, source_fn = os.path.split(source_fp)
        fn, old_ext = os.path.splitext(source_fn)
        if len(old_ext) > 10: # not an extension but part of the file name
            source_fn += self.extension
        else:
            source_fn = fn + self.extension
        if self.dest_folder == "converted":
            self.dest_folder = os.path.join(source_folder, "converted")
            if not os.path.exists(self.dest_folder):
                os.makedirs(self.dest_folder)
        self.dest_fp = os.path.join(self.dest_folder, source_fn)
        return self.dest_fp


    def get_metadata(self, source_fp):
        """Gets the metadata from the file at source_fp.
        (### DUMMY: returns string with magic_value and header_splitter only.
        Overwrite this method in subclass ###)

        Args:
            source_fp (str): path to the original file

        Returns:
            metadata (str): metadata formatted in OpenITI format
                (including the magic value and the header splitter)
        """
        metadata = ""
        metadata = self.magic_value + metadata + self.header_splitter
        return metadata


    def get_data(self, source_fp):
        """Extract the data from the source_fp and format it in OpenITI format.

        ### Overwrite this method in subclass ###

        Args:
            source_fp (str): path to the original file

        Returns:
            text (str): the text in its initial state
        """
        with open(source_fp, mode="r", encoding="utf-8") as file:
            text = file.read()
        return text


    def pre_process(self, text):
        """Remove unwanted features from the text.

        This method removes Arabic vowels and other non-basic characters
        from the text, and normalizes composite letters.

        ### DO NOT overwrite this method in subclass unless it is necessary;
        First, try extending it by using `text = super().pre_process`
        in the subclass's `pre_process` method

        Args:
            text (str): the text in its initial state

        Returns:
            text (str): pre-processed text
        """
        text = deNoise(text)
        text = normalize_composites(text)
        
        repl = [("۱", "١"),   # EXTENDED ARABIC-INDIC DIGIT
                ("۲", "٢"),   # EXTENDED ARABIC-INDIC DIGIT
                ("۳", "٣"),   # EXTENDED ARABIC-INDIC DIGIT
                ("۴", "٤"),   # EXTENDED ARABIC-INDIC DIGIT
                ("۵", "٥"),   # EXTENDED ARABIC-INDIC DIGIT
                ("ک", "ك"),   # Farsi kaf
                ("ی", "ي"),   # Farsi ya
                ("ے", "ي"),   # Farsi yeh barri
                ("‌", ""),     # ZERO WIDTH NON-JOINER
                ("‍", ""),     # ZERO WIDTH JOINER
                ("ۀ", "ة"),   # Farsi ta marbuta in ezafe construction
                ("ۂ", "ة"),   # Farsi ta marbuta in ezafe construction (bis)
                ]
        text = normalize(text, repl)
        
        # attach separated wa- and a- prefixes to the following word: 
        text = re.sub(r"\b([وأ])[\s~]+", r"\1", text)
        return text


    def add_page_numbers(self, text, source_fp):
        """Add the page numbers in OpenITI format to the text.

        ### DUMMY METHOD. Overwrite this method in subclass ###

        Args:
            text (str): the text in its current conversion state
            source_fp (str): the path of the source file
               (sometimes contains information on the volumen number)

        Returns:
            text_w_pages (str): text with the page numbers added
                (in OpenITI format).
        """
        text_w_pages = text
        return text_w_pages


    def add_structural_annotations(self, text):
        """Add strucural annotations in OpenITI format to the text.

        ### DUMMY METHOD. Overwrite this method in subclass ###

        Args:
            text (str): the text in its current conversion state

        Returns:
            text_w_pages (str): text with the structural markup
                (in OpenITI format) added.
        """
        ann_text = text
        return ann_text


    def remove_notes(self, text):
        """Remove the footnotes from the text and convert them into endnotes.

        ### DUMMY METHOD. Overwrite this method in subclass ###

        Args:
            text (str)

        Returns:
            text_without_notes (str): text from which the notes were removed.
            notes (str): the footnotes converted to endnotes
                (if there are endnotes: add the endnote_splitter before them)
        """
        text_without_notes = text
        notes = ""
        if notes:
            notes = self.endnote_splitter + notes
        return text_without_notes, notes


    def reflow(self, text, max_len=None):
        """wrap long lines, i.e. break long lines down \
        by inserting a new line character (and two tildes) \
        so that no line is longer than max_len characters. \
        
        See https://docs.python.org/3.1/library/textwrap.html
        Mark the start of paragraphs with "# ".

        Args:
            text (str): text string to be wrapped
            max_len (int): "width" = max number of characters in a line.
                Default: None (in this case, the default value will be used
                (self.max_line_len)

        Returns:
            newtext (str): wrapped string

        Examples:
            >>> import generic_converter
            >>> g = generic_converter.GenericConverter()
            >>> text = "# This is a short paragraph\\n\\n\
# This is a longer paragraph (longer than max_len)"
            >>> g.reflow(text, max_len=40)
            '# This is a short paragraph\\n\\n\
# This is a longer paragraph (longer\\n\
~~than max_len)'
            
        """
        if max_len:
            self.max_line_len = max_len
        newtext = []

        # split the text, keeping the number of newline characters:

        text = re.sub("\r", "\n", text)
        text = re.split("(\n\n+)", text)

        # wrap each line:

        ignoreTuple = ("#000000#", "#NewRec#", "#####", "### ",
                       "#META#", "Page")
        for line in text:
            if not line.startswith(ignoreTuple):
                if line != "" and "\n\n" not in line:

                    # make sure the new lines that signal the end of a tag
                    # are treated differently from textwrap newlines:
                    
                    sublines = line.split("\n")
                    new_line = []
                    for s in sublines:
                        s_wrap = textwrap.wrap(s, self.max_line_len)
                        new_line.append("\n~~".join(s_wrap))
                    line = "\n".join(new_line)
            newtext.append(line)

        # re-assemble the text:
        
        newtext = "".join(newtext)

        # unwrap lines that are less than 6 characters long:
        
        newtext = re.sub(r"[\r\n]+~~(.{1,6}?[\r\n]+)", r" \1", newtext)
        return newtext


    def add_milestones(self, text):
        """Add a milestone tag after a pre-defined number of words.

        NB: THIS METHOD IS OLD AND MUST BE REPLACED WITH MASOUMEH'S!
        """
##
##        The number of words is defined in the self.chunk_length variable.
##
##        Example:
##            >>> import generic_converter
##            >>> g = generic_converter.GenericConverter()
##            >>> g.chunk_length = 5
##            >>> g.milestone_tag = "###"
##            >>> text = '\
##بِسْمِ اللَّهِ الرَّحْمَنِ الرَّحِيمِ\
## * الْحَمْدُ لِلَّهِ رَبِّ الْعَالَمِينَ\
## * الرَّحْمَنِ الرَّحِيمِ\
## * مَالِكِ يَوْمِ الدِّينِ\
## * إِيَّاكَ نَعْبُدُ وَإِيَّاكَ نَسْتَعِينُ\
## * اهْدِنَا الصِّرَاطَ الْمُسْتَقِيمَ\
## * صِرَاطَ الَّذِينَ أَنْعَمْتَ عَلَيْهِمْ غَيْرِ الْمَغْضُوبِ عَلَيْهِمْ وَلَا الضَّالِّينَ'
##            >>> g.add_milestones(text)
##            '\
##بِسْمِ اللَّهِ الرَّحْمَنِ الرَّحِيمِ\
## * الْحَمْدُ ### لِلَّهِ رَبِّ الْعَالَمِينَ\
## * الرَّحْمَنِ الرَّحِيمِ\
## * ### مَالِكِ يَوْمِ الدِّينِ\
## * إِيَّاكَ نَعْبُدُ ### وَإِيَّاكَ نَسْتَعِينُ\
## * اهْدِنَا الصِّرَاطَ الْمُسْتَقِيمَ\
## * ### صِرَاطَ الَّذِينَ أَنْعَمْتَ عَلَيْهِمْ غَيْرِ ### الْمَغْضُوبِ عَلَيْهِمْ وَلَا الضَّالِّينَ'
##        """
##        # start counting only after the first ### | tag
##        try:
##            editorial, text = text.split("### | ", maxsplit=1)
##            text = "### | " + text
##        except:
##            editorial = ""
##        ms_text = ""  # the text with the milestones added
##        token_count = 0
##        milestones = 0
##        for token in self.ara_tok.finditer(text):
##            #print(token_count, [token.group()])
##            if re.findall(self.ara_char, token.group()[0]):
##                if token_count == self.chunk_length:
##                    token_count = 1
##                    milestones += 1
##                    ms_text += self.milestone_tag
##                else:
##                    token_count += 1
##            else:
##                pass  # do not increment the token count for non-Arabic tokens
##            ms_text += token.group()
##
##        # add spaces before and after the milestone_tag:
##        ms_text = re.sub(" *({}) *".format(self.milestone_tag),
##                         r" \1 ", ms_text)
##
##        return editorial + ms_text
        return text


    def post_process(self, text):
        """Carry out post-processing operations on the text.

        ### Overwrite this method in subclass ###

        Args:
            text (str): the text in its current conversion state

        Returns:
            processed (str): the text after post-processing operations.
        """
        processed = re.sub("[\n\r]{2,}(?![#\|P])", "\n\n# ", text)
        processed = re.sub("[\n\r]+#* *[\n\r]+", "\n\n", processed)
        return processed

    def compose(self, metadata, text, notes):
        """Compose the final text from its parts.

        ### Overwrite this method in subclass if necessary ###

        Args:
            metadata (str): the metadata string in OpenITI format
                (including the magic_value and header_splitter)
            text (str): the main text (incl. page numbers and milestones)
            notes (str): the footnotes, converted to endnotes.

        Returns:
            composite_text (str)
        """
        composite_text = metadata + text + notes
        return composite_text


    def save_file(self, text, dest_fp):
        if not os.path.exists(os.path.split(dest_fp)[0]):
            os.makedirs(os.path.split(dest_fp)[0])
        with open(dest_fp, mode="w", encoding="utf-8") as file:
            file.write(text)


if __name__ == "__main__":
    import doctest
    doctest.testmod()
    print("all test passed")

    input("Press Enter to start converting")

    conv = GenericConverter()
    conv.convert_file("test/test.txt")







"""Convert Epubs from the Hindawi library to OpenITI mARkdown.

The converter has two main functions:
* convert_file: convert a single epub file.
* convert_files_in_folder: convert all epub files in a given folder

Usage examples:
    >>> folder = r"test/hindawi/"
    >>> meta_fp = folder+"hindawi_metadata_man.yml"
    >>> from epub_converter_hindawi import convert_file, convert_files_in_folder
    >>> src_fp = folder+"26362727.epub"
    >>> convert_file(src_fp, meta_fp, dest_fp=folder+"converted/26362727")
    >>> convert_files_in_folder(folder, meta_fp, dest_folder=folder+"converted")
    Converting all files in folder test/hindawi/ with extensions ['epub']

Both functions use the HindawiEpubConverter class to do the heavy lifting.
The HindawiEpubConverter is a subclass of the GenericEpubConverter,
which in turn is a subclass of the GenericConverter
from the generic_converter module:

GenericConverter
    \_ GenericEpubConverter
            \_ HindawiEpubConverter

Methods of both classes:

(methods of GenericConverter are inherited by GenericEpubConverter;
methods of GenericConverter with the same name
in GenericEpubConverter are overwritten by the latter)

=========================== ========================= =======================
generic_converter           epub_converter_generic    epub_converter_hindawi 
=========================== ========================= =======================
__init__                    __init__                  __init__ 
convert_files_in_folder     (inherited)               (inherited)
convert file                (inherited)               (inherited)
make_dest_fp                (inherited - generic!)    (inherited - generic!)
get_metadata                (inherited - generic!)    get_metadata
get_data                    get_data                  (inherited)
pre_process                 (inherited)               (inherited)
add_page_numbers            (inherited - generic!)    (inherited - generic!)
add_structural_annotations  (inherited - generic!)    (inherited - generic!) 
remove_notes                remove_notes              (inherited)
reflow                      (inherited)               (inherited)
add_milestones              (inherited)               (inherited)
post_process                (inherited - generic!)    (inherited - generic!) 
compose                     (inherited)               (inherited)
save_file                   (inherited)               (inherited)
                            convert_html2md           convert_html2md
                            inspect_epub              (inherited)
                            sort_html_files_by_toc    (inherited)
                            add_unique_tags           (inherited)
=========================== ========================= =======================


Examples:
    >>> from epub_converter_hindawi import HindawiEpubConverter
    >>> from helper.yml2json import yml2json
    >>> folder = "test/"
    >>> fn = "26362727.epub"
    >>> hc = HindawiEpubConverter(dest_folder="test/converted")
    >>> hc.VERBOSE = False
    >>> meta_fp = "test/hindawi/hindawi_metadata_man.yml"
    >>> hc.metadata_dic = yml2json(meta_fp, container = {})
    >>> hc.metadata_file = meta_fp
    >>> hc.convert_file(folder+fn)

    #>>> hc.convert_files_in_folder(folder)

"""

import os

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(root_folder))
    sys.path.append(root_folder)

from openiti.new_books.convert.epub_converter_generic import GenericEpubConverter
from openiti.new_books.convert.helper import html2md_hindawi
from openiti.new_books.convert.helper.yml2json import yml2json

##from epub_converter_generic import GenericEpubConverter
##import html2md_hindawi
##from yml2json import yml2json


def convert_file(fp, meta_fp, dest_fp=None, verbose=False):
    """Convert one file to OpenITI format.

    Args:
        fp (str): path to the file that must be converted.
        meta_fp (str): path to the yml file containing the Hindawi metadata
        dest_fp (str): path to the converted file.

    Returns:
        None
    """
    conv = HindawiEpubConverter()
    conv.VERBOSE = verbose
    conv.metadata_dic = yml2json(meta_fp, container = {})
    conv.metadata_file = meta_fp
    conv.convert_file(fp, dest_fp=dest_fp)

def convert_files_in_folder(src_folder, meta_fp, dest_folder=None, verbose=False,
                            extensions=["epub"], exclude_extensions=["yml"],
                            fn_regex=None):
    """Convert all files in a folder to OpenITI format.\
    Use the `extensions` and `exclude_extensions` lists to filter\
    the files to be converted.

    Args:
        src_folder (str): path to the folder that contains
            the files that must be converted.
        meta_fp (str): path to the yml file containing the Hindawi metadata
        dest_folder (str): path to the folder where converted files
            will be stored.
        extensions (list): list of extensions; if this list is not empty,
            only files with an extension in the list should be converted.
        exclude_extensions (list): list of extensions;
            if this list is not empty,
            only files whose extension is not in the list will be converted.
        fn_regex (str): regular expression defining the filename pattern
            e.g., "-(ara|per)\d". If `fn_regex` is defined,
            only files whose filename matches the pattern will be converted.

    Returns:
        None
    """
    msg = "Converting all files in folder {} with extensions {}"
    print(msg.format(src_folder, extensions))
    conv = HindawiEpubConverter()
    conv.VERBOSE = verbose
    conv.metadata_dic = yml2json(meta_fp, container = {})
    conv.metadata_file = meta_fp
    conv.convert_files_in_folder(src_folder, dest_folder=dest_folder,
                                 extensions=extensions,
                                 exclude_extensions=exclude_extensions,
                                 fn_regex=fn_regex)


################################################################################




class HindawiEpubConverter(GenericEpubConverter):
    def __init__(self, dest_folder=None):
        super().__init__(dest_folder)
        self.toc_fn = "nav.xhtml"
        self.metadata_file = None

    def convert_html2md(self, html):
        """Use custom html to mARKdown function for Hindawi epubs."""
        text = html2md_hindawi.markdownify(html)
        return text

    def get_metadata(self, metadata_fp):
        """Custom method to get the metadata of the Hindawi epub file."""
        source_fp = self.source_fp
        bookID = os.path.split(source_fp)[1]
        bookID = os.path.splitext(bookID)[0]
        meta_dic = self.metadata_dic[bookID]
        meta = ["#META# {}: {}".format(k,v) for k,v in sorted(meta_dic.items())]
        return self.magic_value + "\n".join(meta) + self.header_splitter


if __name__== "__main__":
    import doctest
    doctest.testmod()
    input("Testing finished. Continue?")

    hc = HindawiEpubConverter(dest_folder="test/converted")

    # identify the location of the yml file containing the metadata:
    meta_fp = r"test\hindawi_metadata_man.yml"
    hc.metadata = yml2json(meta_fp, container={})
    
    fp = r"test\26362727.epub"
    hc.convert_file(fp)
    print("converted Hindawi epub", fp)

    hc.convert_files_in_folder("test/hindawi")
    print("converted all epub files in folder", "test/hindawi")




"""Convert LAL library html to OpenITI mARkdown.

This script subclasses the generic MarkdownConverter class
from the html2md module (based on python-markdownify,
https://github.com/matthewwithanm/python-markdownify),
which uses BeautifulSoup to create a flexible converter.
The subclass in this module, LALHtmlConverter,
adds methods specifically for the conversion of books from
the LAL library to OpenITI mARkdown:

* span conversion: the LAL html seems to be a conversion of tei xml;
    the tei data is often embedded inside the id of a span.

Inheritance schema of the LALHtmlConverter:

======================== ==========================
MarkdownConverter        LALHtmlConverter
======================== ==========================
Options                  (inherited)
DefaultOptions           (inherited)
__init__                 (inherited)
__getattr__              (inherited)
convert                  (inherited)
process_tag              (inherited)
process_text             (inherited)
fill_out_columns         (inherited)
post_process_md          post_process_md (appended)
should_convert_tag       (inherited)
indent                   (inherited)
underline                (inherited)
create_underline_line    (inherited)
convert_a                (inherited)
convert_b                (inherited)
convert_blockquote       (inherited)
convert_br               (inherited)
convert_em               (inherited)
convert_hn               (inherited)
convert_i                (inherited)
convert_img              (inherited)
convert_list             (inherited)
convert_li               (inherited)
convert_ol               (inherited)
convert_p                convert_p
convert_table            (inherited)
convert_tr               (inherited)
convert_ul               (inherited)
convert_strong           (inherited)
                         convert_div
                         convert_opener
                         convert_closer
                         convert_cit
                         convert_l
======================== ==========================

"""
import re

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(path.dirname(root_folder)))
    sys.path.append(root_folder)

from openiti.new_books.convert.helper import html2md
from openiti.new_books.convert.helper.html2md import *  # import all constants!
from openiti.helper.quran import normalise_quran, special_letters_regex

class LALHtmlConverter(html2md.MarkdownConverter):
    """Convert LAL library html to OpenITI mARkdown."""

##    def __init__(self, **options):
##        super().__init__(**options)
##        self.class_dict = dict()
##        self.class_dict["linebreak"] = '\n'

    def convert(self, html):
        """Convert XML to markdown.

        NB: replaces the parent class's convert method because it uses an html parser rather than xml!

        # We want to take advantage of the html5 parsing, but we don't actually
        # want a full document. Therefore, we'll mark our fragment with an id,
        # create the document, and extract the element with the id.
        """

        html = wrapped % html
        #soup = BeautifulSoup(html, 'html.parser')
        soup = BeautifulSoup(html, 'xml')
        if 'strip' in self.options and self.options["strip"]:
            for tag in self.options["strip"]:
                [t.decompose() for t in soup.find_all(tag)]
        text = self.process_tag(soup.find(id=FRAGMENT_ID), children_only=True)

##        # post-processing: remove unneeded blank lines and spaces:
##
##        # remove leading and trailing spaces in lines:
##        text = re.sub(r" *(\n+) *", r"\1", text)
##        # remove unwanted additional spaces and lines:
##        text = re.sub(r"\n{3,}", r"\n\n", text)
##        text = re.sub(r" +", r" ", text)
##        # remove blank lines marked with "DELETE_PREVIOUS_BLANKLINES" tag
##        text = re.sub(r"\n+DELETE_PREVIOUS_BLANKLINES", "", text)
##        # replace placeholders for spaces in tables:
##        text = re.sub("ç", " ", text)
##        return text
        return self.post_process_md(text)

    def convert_div(self, el, text):
        """Convert div tags (structural markup)"""
        # def alpha2num(s):
        #     """Convert alphabetical part of an ID to numerical"""
        #     # this approach doesn't work for things like 1.b.2!
        #     alpha = re.findall("[a-zA-Z]+$", s) 
        #     # no conversion needed if the ID does not contain letters:
        #     if not alpha:
        #         return s
        #     # calculate the numerical value of an alphabetic string (a=1, aa=27, ...)
        #     alphabet = "abcdefghijklmnopqrstuvwxyz"
        #     num = [alphabet.index(c.lower())+1 for c in alpha[0]]
        #     sum_num = 0
        #     for i,n in enumerate(num[::-1]):
        #         sum_num += (26**i)*n
        #     return s[:-len(alpha[0])] + str(sum_num)

        def alpha2num(s):
            """Convert alphabetical part of an ID to numerical"""
            alpha = re.findall("[a-zA-Z]", s) 
            # no conversion needed if the ID does not contain letters:
            if not alpha:
                return s
            # calculate the numerical value of an alphabetic string (a=1, aa=27, ...)
            alphabet = "abcdefghijklmnopqrstuvwxyz"
            num_s = []
            for el in s.split("."):
                if re.findall("[a-zA-Z]", el):
                    num = [alphabet.index(c.lower())+1 for c in el]
                    sum_num = 0
                    for i,n in enumerate(num[::-1]):
                        sum_num += (26**i)*n
                    num_s.append(str(sum_num))
                else:
                    num_s.append(el)
            return ".".join(num_s)

        if "type" in el.attrs and "prelim" in el["type"]:
            return "\n### |PARATEXT|" + text
        
        # use xml:id (e.g., xml:id="ARA.section.4.1.2") to get the section number
        xml_id = el["xml:id"]
        xml_id = re.sub("\.(\d+)([a-zA-Z]+)$", r".\1.\2", xml_id)            
        #n = ".".join(xml_id.split(".")[2:])
        # convert the alphabetical parts of a section ID (e.g. 4.1.b) to numbers (e.g., 4.1.2)
        #n = re.findall("(?:\d+\.)*[\da-zA-Z]+$", xml_id)[0] # this approach does not work for things like 4.b.2...
        try:
            n = ".".join(xml_id.split(".")[2:])
            n = alpha2num(n)
        except Exception as e:
            print(xml_id, e)
            n = ""
        # calculate the hierarchical level of the header (that is, the number of pipes in the ### |+ tag)
        level = len(n.split("."))
        header = ""
        for i, c in enumerate(el.findChildren()):
            if c.name == "head":
                header += " " + c.text.strip()
            elif c.name in ("div", "p") or i > 4:
                break

        if "type" in el.attrs and el["type"] != "section":
            #return "\n### {} {}{}".format(level*"|", header, text)
            #return "\n### {} {}{}{}".format(level*"|", n.split(".")[-1], header, text)
            #return "\n### {} {}{}{}".format(level*"|", n, header, text)
            if header:
                header_regex = re.sub("\s+", r"\\s*", header)
                text = re.sub(header_regex, "", text, count=1)
                return "\n### {} {} {}\n{}".format(level*"|", n, header, text)
            return "\n### {} {} {}".format(level*"|", n, text)
        else:
            return "\n[{}]{}".format(n, text)
        
    def convert_l(self, el, text):
        """Convert poetry line tags"""
        # caesuras are dealt with in preprocessing
        #for caesura in tag.find_all("caesura"):
        #    caesura.replace_with(" %~% ")
        #line = "\n# " + tag.text.strip()
        line = "\n# " + text
        if not " %~% " in line:
            line += " %~%"
        return line

    def convert_opener(self, el, text):
        """Convert opener tag (paratext)"""
        #return "\n### |PARATEXT|\n# " + tag.text.strip()
        #return "\n### |PARATEXT|\n# " + text
        return "\n# " + text  # most often, not really paratext.

    def convert_closer(self, el, text):
        """Convert closer tag (paratext)"""
        #return "\n### |PARATEXT|\n# " + tag.text.strip()
        if "rend" in el.attrs and "ornament" in el["rend"]:
            return "\n" + text
            
        return "\n### |PARATEXT|\n# " + text

    def convert_cit(self, el, text):
        """Convert cit tag (for Qur'an citations)"""
        quote = []
        for q in el.find_all("quote"):
            quote.append(q.text.strip())
        quote = " * ".join(quote)
        quote = normalise_quran(quote, post_process=True)
        qid = []
        try:    
            for pointer in el.find_all("ptr"):
                qid.append(re.sub("Q\W*", "", pointer["cref"]))
        except Exception as e:
            pass
        if qid:
            #return " @QB@{}@ {} @QE@{}@ ".format(qid, el.quote.text.strip(), qid)
            qid = "_".join(qid)
            return " @QUR_{}@ {}\n".format(qid, quote)
        #return " @QB@ {} @QE@ ".format(el.quote.text.strip())
        return " @QUR@ {}\n".format(quote)

    def convert_item(self, el, text):
        """Convert list items"""
        return "\n# * " + text

    def convert_p(self, el, text):
        """Convert p tags (paragraphs)"""
        if "n" in el.attrs:
            return "\n# " + el["n"] + " " + text
        if "rend" in el.attrs:
            if "isnad" in el["rend"]:
                #return "\n# @ISNAD@ "+tag.text.strip()
                return "\n# _ISNAD_ " + text
            elif "answer" in el["rend"]:
                pass
            elif "question" in el["rend"]:
                pass
            elif "bismillah" in el["rend"]:
                pass
            elif "no_indent"in el["rend"]:
                pass
            elif "tafsir" in el["rend"]:
                pass
            
        #return "\n# "+tag.text.strip()
        return "\n# " + text

    def post_process_md(self, text):
        """Appends to the MarkdownConverter.post_process_md() method."""
        text = re.sub(" *\n~~ *\n", "\n", text)
        text = re.sub("(### \|PARATEXT\|(?:\n# .+)+)\n### \|PARATEXT\|", r"\1", text)
        text = super().post_process_md(text)
        text = re.sub("@QUR_", "@QUR@", text)
        # fix duplicate headings:
        text = re.sub(r"(### \|+ [\w.]+) ?[\r\n]+\1", r"\1", text) # number only first
        text = re.sub(r"(### \|+ [\w.]+)( .+)[\r\n]+\1 ?[\r\n]+", r"\1\2\n", text) # header first
        
        return text


def markdownify(html, **options):
    """Shortcut to the convert method of the HindawiConverter class."""
    return LALHtmlConverter(**options).convert(html)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Replaces regular expression patterns in OpenITI texts.
"""
import os
import re
import sys

if __name__ == '__main__':
    main_folder = input("Enter the path to the folder containing OpenITI texts: ")
    overwrite = input("Do you want to overwrite the files (y/n)? ")
    # pat_type = input("Remove only OpenITI mARkdown (y) or a specific regular expression patterns (n)? ")
    pats = {}
    # if pat_type.lower() == "y":
    #     rm_pats.append(" *### \|+ ")
    # elif pat_type.lower() == "n":
    conti = "y"
    while conti.lower() == 'y':
        pat = input("Enter the custom pattern to find: ")  #" *### \|+ "
        replace = input("Enter the custom pattern to replace: ")
        pats[pat] = replace
        conti = input("more patterns to remove (y/n)? ").lower()
    # page_nr_pat = "PageV\d+P\d+"

    if len(sys.argv) > 0:
        if not os.path.exists(main_folder):
            print("invalid path: ", main_folder)
        else:
            # process all texts
            for root, dirs, files in os.walk(main_folder):
                for file in files:
                    if re.search("^\d{4}\w+\.\w+\.\w+-[a-z]{3}\d{1}(\.(mARkdown|inProgress|completed))?$", file):
                        path_full = os.path.join(root, file)
                        with open(path_full, "r", encoding="utf8") as f1:
                            text = f1.read()
                            # remove the specified patterns in the texts
                            for pat in pats:
                                text = re.sub(pat, pats[pat], text)
                        if overwrite.lower() == "y":
                            write_path = path_full
                        else:
                            write_path = path_full + "_replaced"
                        with open(write_path, "w", encoding="utf8") as f9:
                            f9.write(text)



import setuptools
import openiti

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

##################################################################

#    MAKE SURE YOU CHANGED THE VERSION IN OPENITI.__INIT__.PY !

##################################################################

#print("Make sure you have changed the version number in openiti.__init__.py")
#input("Press Enter to continue. ")

setuptools.setup(
    name="openiti",
    version=openiti.__version__,
    author="Sohail Merchant, Maxim Romanov, Masoumeh Seydi, Peter Verkinderen",
    author_email="peter.verkinderen@gmail.com",
    description="A package for dealing with the openITI corpus",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/OpenITI",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "beautifulsoup4",
        "requests",
        "six",
        "sphinx_rtd_theme",
        # added the following three to solve Read The Docs import problem:
        "pypyodbc",
        "bs4",
        "pygithub",
    ],
    python_requires='>=3.4',
)

